﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Clio.XmlEngine;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using TreeSharp;
using Action = TreeSharp.Action;
using Buddy.Coroutines;
using ff14bot.Enums;
using ff14bot.Objects;

namespace ff14bot.NeoProfiles {
    [XmlElement("Craft3Start40Dura")]
    class Craft3Start40DuraTag : ProfileBehavior {
        private bool _done = false;
        public override bool IsDone { get { return _done; } }

        private Composite _cache;
        private static Random rng = new Random();

        // Same function as MinDelay from CraftAction, however this will apply to all actions taken
        [XmlAttribute("MinDelay")]
        public int MinDelay { get; set; }
        // Same function as MinDelay from CraftAction, however this will apply to all actions taken
        [XmlAttribute("MaxDelay")]
        public int MaxDelay { get; set; }


        #region counters
        protected bool FinisherStarted = false;
        protected int RapidsUsed = 0;
        protected int ManipulationsUsed = 0;
        protected int ComfortZonesUsed = 0;
        protected int TricksUsed = 0;

        protected void Reset() {
            FinisherStarted = false;
            RapidsUsed = 0;
            ManipulationsUsed = 0;
            ComfortZonesUsed = 0;
            TricksUsed = 0;
        }
        #endregion

        Composite Behavior() {
            return new PrioritySelector(
                new Decorator(ret => !CraftingManager.IsCrafting,
                    new Action(r => {
                        _done = true;
                    })
                ),
                // don't waste any manipulation stacks
                new Decorator(ret => ActionsRemaining == 4 && ManipulationRemaining > 0, new ActionRunCoroutine(ctx => MainAction())),
                new Decorator(ret => Condition == Good, TRICKS_OF_THE_TRADE),
                new Decorator(ret => ShouldIFinish(), new ActionRunCoroutine(ctx => Finisher())),
                new Decorator(ret => ShouldIComfortZone(), COMFORT_ZONE),
                new Decorator(ret => InnerQuietStacks == 0, INNER_QUIET),
                new Decorator(ret => ShouldIManipulation(), MANIPULATION),
                new Decorator(ret => ShouldISteadyHandII(), STEADY_HAND_II),
                new Decorator(ret => ActionsRemaining > 1, new ActionRunCoroutine(ctx => MainAction()))
            );
        }

        public bool ShouldIFinish() {
            if (ActionsRemainingWithManipulation != 3) return false;

            // need 166 for a Manipulation + 78CP finisher
            // set it to 140 since there's a fairly good chance we'll get another tott with another manipulation
            if (ManipulationsUsed == 3 && CPInSteps(5) < 146) return true;

            if (ManipulationsUsed == 4) return true;

            return false;
        }

        public bool ShouldISteadyHandII() {
            if (SteadyHandIIRemaining > 0) return false;

            if (ManipulationsUsed < 4) return true;

            if (ManipulationsUsed == 4 && CPInSteps(4) > 78) return true;

            // somehow got to a point where we're down to 10 durability
            // with manipulation up, so need to SHII to let Manipulation tick once.
            if (ActionsRemaining == 1 && ManipulationRemaining > 0) return true;

            return false;
        }

        public bool ShouldIManipulation() {
            if (ManipulationRemaining > 0) return false;

            if (ActionsRemaining > 2) return false;

            // we don't need to use manipulation until 10 durability left since refreshing SHII will bring us back up from 10
            if (ActionsRemaining == 2 && SteadyHandIIRemaining > 2) return true;

            // we better follow this up with something!
            if (ActionsRemaining == 1) return true;

            return false;
        }

        /**
         * Not happy about this implementation for whether or not to use the 3rd CZ.
         * I've had instances where I'll start the finisher with 5 stacks of CZ.
         * 
         * May want to consider using # of ToTT procs instead.
         * */
        public bool ShouldIComfortZone() {
            if (ComfortZoneRemaining > 0) return false;
            
            if (ComfortZonesUsed < 2) return true;

            // make sure we have enough cp for manipulation #4
            if (ComfortZonesUsed == 2) {
                if (Core.Me.CurrentCP >= 146 && ManipulationsUsed < 3)
                    return true;

                // safe to use another CZ if we've got at least 3 stacks of Manipulation left?
                if (ManipulationsUsed == 3 && ManipulationRemaining >= 3)
                    return true;
            }

            return false;
        }

        /**
         * (96) SHI > (74) GS > (42) Innovation > (24) BB > CSII > CSII
         * (78) SHI > (56) GS > (24) BB > CSII > CSII
         * (46) SHI > (24) BB > CSII > CSII
         */
        public async Task<bool> Finisher() {
            // in the rare instance that we get an Excellent right at the start
            // and already have SHII from the main rotation, just hit BB and go
            if (Condition == Excellent && SteadyHandIIRemaining > 0) {
                await CraftAction("Byregot's Blessing");
                await CraftAction("Careful Synthesis II");
                await CraftAction("Careful Synthesis II");
                return true;
            }


            // if rotation ended with a 4+ stack of SHII, no need to refresh
            if (SteadyHandIIRemaining < 4) {
                await CraftAction("Steady Hand");
                if (Condition == Good) {
                    await CraftAction("Tricks of the Trade");
                }
            }

            if (Condition == Excellent) {
                await CraftAction("Byregot's Blessing");
                await CraftAction("Careful Synthesis II");
                await CraftAction("Careful Synthesis II");
                return true;
            }

            if (CPInSteps(1) >= 56)
                await CraftAction("Great Strides");

            if (Condition == Excellent || Condition == Good) {
                await CraftAction("Byregot's Blessing");
                await CraftAction("Careful Synthesis II");
                await CraftAction("Careful Synthesis II");
                return true;
            }

            if (CPInSteps(1) >= 42)
                await CraftAction("Innovation");

            if (Condition == Excellent || Condition == Good) {
                await CraftAction("Byregot's Blessing");
                await CraftAction("Careful Synthesis II");
                await CraftAction("Careful Synthesis II");
                return true;
            }

            // use a CSII before the BB to fish for a Good/Excellent.
            await CraftAction("Careful Synthesis II");
            await CraftAction("Byregot's Blessing");
            await CraftAction("Careful Synthesis II");
            return true;
        }

        public async Task<bool> MainAction() {
            if (RapidsUsed == 1) {
                await CraftAction("Hasty Touch");
            } else {
                await CraftAction("Rapid Synthesis");
            }

            return true;
        }

        #region Base Methods
        public async Task<bool> CraftAction(string ActionName) {
            SpellData Action = GetAction(ActionName);
            await Coroutine.Wait(Timeout.Infinite, () => !CraftingManager.AnimationLocked);
            int StartingProgress = CraftingManager.Progress;

            if (Actionmanager.CanCast(Action, null)) {
                Logging.Write("Step{0}. {1} :: {2} :: {3}/{4}CP :: {5}/{6}D :: {7}/{8}Q :: {9}/{10}P :: {11}IQ :: {12}-{13}CZ :: {14}SHII :: {15}-{16}M",
                    Step,
                    Action.LocalizedName,
                    CraftingManager.Condition.ToString(),
                    CP, Core.Me.MaxCP,
                    CraftingManager.Durability, CraftingManager.DurabilityCap,
                    CraftingManager.Quality, CraftingManager.QualityCap,
                    CraftingManager.Progress, CraftingManager.ProgressRequired,
                    InnerQuietStacks,
                    ComfortZoneRemaining, ComfortZonesUsed,
                    SteadyHandIIRemaining,
                    ManipulationRemaining, ManipulationsUsed
                    );
                Actionmanager.DoAction(Action, null);
                if (ActionName == "Manipulation") ManipulationsUsed++;
                if (ActionName == "Comfort Zone") ComfortZonesUsed++;
                if (ActionName == "Tricks of the Trade") TricksUsed++;
            } else {
                Logging.Write("Could not cast {0} ({1})", Action.LocalizedName, Action.Id);
                return false;
            }

            await Coroutine.Wait(10000, () => CraftingManager.AnimationLocked);
            await Coroutine.Wait(Timeout.Infinite, () => !CraftingManager.AnimationLocked);

            if (CraftingManager.Progress > StartingProgress && ActionName == "Rapid Synthesis") {
                RapidsUsed++;
            }

            await Coroutine.Sleep(rng.Next(MinDelay, MaxDelay));

            return true;
        }

        ActionRunCoroutine COMFORT_ZONE;
        ActionRunCoroutine INNER_QUIET;
        ActionRunCoroutine STEADY_HAND_II;
        ActionRunCoroutine BASIC_TOUCH;
        ActionRunCoroutine MASTERS_MEND;
        ActionRunCoroutine STEADY_HAND;
        ActionRunCoroutine INGENUITY_II;
        ActionRunCoroutine STANDARD_TOUCH;
        ActionRunCoroutine GREAT_STRIDES;
        ActionRunCoroutine BYREGOTS_BLESSING;
        ActionRunCoroutine INNOVATION;
        ActionRunCoroutine CAREFUL_SYNTHESIS_II;
        ActionRunCoroutine MANIPULATION;
        ActionRunCoroutine TRICKS_OF_THE_TRADE;

        private bool _init = false;
        private void init() {
            if (_init) return;
            _init = true;

            MinDelay = 250;
            MaxDelay = 500;

            COMFORT_ZONE = new ActionRunCoroutine(ctx => CraftAction("Comfort Zone"));
            INNER_QUIET = new ActionRunCoroutine(ctx => CraftAction("Inner Quiet"));
            STEADY_HAND_II = new ActionRunCoroutine(ctx => CraftAction("Steady Hand II"));
            BASIC_TOUCH = new ActionRunCoroutine(ctx => CraftAction("Basic Touch"));
            MASTERS_MEND = new ActionRunCoroutine(ctx => CraftAction("Master's Mend"));
            STEADY_HAND = new ActionRunCoroutine(ctx => CraftAction("Steady Hand"));
            INGENUITY_II = new ActionRunCoroutine(ctx => CraftAction("Ingenuity II"));
            STANDARD_TOUCH = new ActionRunCoroutine(ctx => CraftAction("Standard Touch"));
            GREAT_STRIDES = new ActionRunCoroutine(ctx => CraftAction("Great Strides"));
            BYREGOTS_BLESSING = new ActionRunCoroutine(ctx => CraftAction("Byregot's Blessing"));
            INNOVATION = new ActionRunCoroutine(ctx => CraftAction("Innovation"));
            CAREFUL_SYNTHESIS_II = new ActionRunCoroutine(ctx => CraftAction("Careful Synthesis II"));
            MANIPULATION = new ActionRunCoroutine(ctx => CraftAction("Manipulation"));
            TRICKS_OF_THE_TRADE = new ActionRunCoroutine(ctx => CraftAction("Tricks of the Trade"));
        }
        #endregion

        #region Helper Getters
        public int Step {
            get {
                return CraftingManager.Step;
            }
        }
        public int CP {
            get {
                return Core.Me.CurrentCP;
            }
        }
        public int CPInSteps(int Steps) {
            int Max = Math.Min(Steps, ComfortZoneRemaining);

            return CP + Max * 8;
        }
        public CraftingCondition Condition {
            get {
                return CraftingManager.Condition;
            }
        }
        public CraftingCondition Poor = CraftingCondition.Poor;
        public CraftingCondition Normal = CraftingCondition.Normal;
        public CraftingCondition Good = CraftingCondition.Good;
        public CraftingCondition Excellent = CraftingCondition.Excellent;

        public int ActionsRemaining {
            get {
                //Logging.Write("ActionsRemaining Check -- {0}", CraftingManager.Durability);
                return (CraftingManager.Durability + 5) / 10;
            }
        }

        public int ActionsRemainingWithManipulation {
            get {
                return (CraftingManager.Durability + 5) / 10 + ManipulationRemaining;
            }
        }

        public int SteadyHandIIRemaining {
            get {
                return GetAuraStacksById(262);
            }
        }

        public int ManipulationRemaining {
            get {
                return GetAuraStacksById(258);
            }
        }

        public int ComfortZoneRemaining {
            get {
                return GetAuraStacksById(261);
            }
        }

        public int InnerQuietStacks {
            get {
                return GetAuraStacksById(251);
            }
        }

        public static int GetAuraStacksById(uint BuffId) {
            if (Core.Me.CharacterAuras.FirstOrDefault(r => r.Id == BuffId) != null) {
                return (int)Core.Me.CharacterAuras.First(r => r.Id == BuffId).Value;
            }

            return 0;
        }
        #endregion

        /*
        // todo
        static int QualityGain(int BaseControl, double Efficiency = 1.0, CraftingCondition Condition = CraftingCondition.Normal, int InnerQuietStacks = 0, bool Innovation = false, bool GreatStrides = false, bool Ingenuity2 = false, bool Ingenuity = false) {
            if (InnerQuietStacks == 0) InnerQuietStacks++;
            double EffectiveControl = BaseControl + BaseControl * (InnerQuietStacks - 1) * 0.2 + (Innovation ? 1 : 0) * .5 * BaseControl;

            double Gain = EffectiveControl * .36 + 34;
            

            return 0;
        }
         */

        protected override void OnResetCachedDone() {
            Reset();
            _done = false;
        }

        protected override void OnDone() {
            TreeHooks.Instance.RemoveHook("TreeStart", _cache);
        }

        protected override void OnStart() {
            init();
            _cache = Behavior();
            TreeHooks.Instance.InsertHook("TreeStart", 0, _cache);
        }

        #region ActionId Resolution
        static SpellData GetAction(string ActionName) {
            uint ActionId = 0;
            if (crossClassActions.ContainsKey(ActionName)) {
                ActionId = crossClassActions[ActionName];
            } else if (classActions.ContainsKey(Core.Player.CurrentJob) && classActions[Core.Player.CurrentJob].ContainsKey(ActionName)) {
                ActionId = classActions[Core.Player.CurrentJob][ActionName];
            }

            return DataManager.GetSpellData(ActionId);
        }

        static Dictionary<ClassJobType, Dictionary<string, uint>> classActions = new Dictionary<ClassJobType, Dictionary<string, uint>>() {
            {ClassJobType.Carpenter, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100001},
                {"Basic Touch", 100002},
                {"Master's Mend", 100003},
                {"Steady Hand", 244},
                {"Inner Quiet", 252},
                {"Observe", 100010},
                {"Rumination", 276},
                {"Standard Touch", 100004},
                {"Great Strides", 260},
                {"Master's Mend II", 100005},
                {"Standard Synthesis", 100007},
                {"Brand of Wind", 100006},
                {"Advanced Touch", 100008},
                {"Byregot's Blessing", 100009}
            }},{ClassJobType.Blacksmith, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100015},
                {"Basic Touch", 100016},
                {"Master's Mend", 100017},
                {"Steady Hand", 245},
                {"Inner Quiet", 253},
                {"Observe", 100023},
                {"Ingenuity", 277},
                {"Standard Touch", 100018},
                {"Great Strides", 261},
                {"Master's Mend II", 100019},
                {"Standard Synthesis", 100021},
                {"Brand of Fire", 100020},
                {"Advanced Touch", 100022},
                {"Ingenuity II", 283}
            }},{ClassJobType.Armorer, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100030},
                {"Basic Touch", 100031},
                {"Master's Mend", 100032},
                {"Steady Hand", 246},
                {"Inner Quiet", 254},
                {"Observe", 100040},
                {"Rapid Synthesis", 100033},
                {"Standard Touch", 100034},
                {"Great Strides", 262},
                {"Master's Mend II", 100035},
                {"Standard Synthesis", 100037},
                {"Brand of Ice", 100036},
                {"Advanced Touch", 100038},
                {"Piece by Piece", 100039}
            }},{ClassJobType.Goldsmith, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100075},
                {"Basic Touch", 100076},
                {"Master's Mend", 100077},
                {"Steady Hand", 247},
                {"Inner Quiet", 255},
                {"Observe", 100082},
                {"Manipulation", 278},
                {"Standard Touch", 100078},
                {"Great Strides", 263},
                {"Master's Mend II", 100079},
                {"Standard Synthesis", 100080},
                {"Flawless Synthesis", 100083},
                {"Advanced Touch", 100081},
                {"Innovation", 284}
            }},{ClassJobType.Leatherworker, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100045},
                {"Basic Touch", 100046},
                {"Master's Mend", 100047},
                {"Steady Hand", 249},
                {"Inner Quiet", 257},
                {"Observe", 100053},
                {"Waste Not", 279},
                {"Standard Touch", 100048},
                {"Great Strides", 265},
                {"Master's Mend II", 100049},
                {"Standard Synthesis", 100051},
                {"Brand of Earth", 100050},
                {"Advanced Touch", 100052},
                {"Waste Not II", 285}
            }},{ClassJobType.Weaver, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100060},
                {"Basic Touch", 100061},
                {"Master's Mend", 100062},
                {"Steady Hand", 248},
                {"Inner Quiet", 256},
                {"Observe", 100070},
                {"Careful Synthesis", 100063},
                {"Standard Touch", 100064},
                {"Great Strides", 264},
                {"Master's Mend II", 100065},
                {"Standard Synthesis", 100067},
                {"Brand of Lightning", 100066},
                {"Advanced Touch", 100068},
                {"Careful Synthesis II", 100069}
            }},{ClassJobType.Alchemist, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100090},
                {"Basic Touch", 100091},
                {"Master's Mend", 100092},
                {"Steady Hand", 250},
                {"Inner Quiet", 258},
                {"Observe", 100099},
                {"Tricks of the Trade", 100098},
                {"Standard Touch", 100093},
                {"Great Strides", 266},
                {"Master's Mend II", 100094},
                {"Standard Synthesis", 100096},
                {"Brand of Water", 100095},
                {"Advanced Touch", 100097},
                {"Comfort Zone", 286}
            }},{ClassJobType.Culinarian, new Dictionary<string, uint>() {
                {"Basic Synthesis", 100105},
                {"Basic Touch", 100106},
                {"Master's Mend", 100107},
                {"Steady Hand", 251},
                {"Inner Quiet", 259},
                {"Observe", 100113},
                {"Hasty Touch", 100108},
                {"Standard Touch", 100109},
                {"Great Strides", 267},
                {"Master's Mend II", 100110},
                {"Standard Synthesis", 100111},
                {"Steady Hand II", 281},
                {"Advanced Touch", 100112},
                {"Reclaim", 287}
            }}
        };

        static Dictionary<string, uint> crossClassActions = new Dictionary<string, uint>() {
            {"Rumination", 276},
            {"Byregot's Blessing", 100009},
            {"Brand of Wind", 100006},
            {"Ingenuity II", 283},
            {"Ingenuity", 277},
            {"Brand of Fire", 100020},
            {"Rapid Synthesis", 100033},
            {"Piece by Piece", 100039},
            {"Brand of Ice", 100036},
            {"Waste Not", 279},
            {"Waste Not II", 285},
            {"Brand of Earth", 100050},
            {"Careful Synthesis", 100063},
            {"Careful Synthesis II", 100069},
            {"Brand of Lightning", 100066},
            {"Tricks of the Trade", 100098},
            {"Brand of Water", 100095},
            {"Comfort Zone", 286},
            {"Hasty Touch", 100108},
            {"Reclaim", 287},
            {"Steady Hand II", 281},
            {"Manipulation", 278},
            {"Innovation", 284},
            {"Flawless Synthesis", 100083}
        };
        #endregion
    }
}
